package com.huawei.dli.jdbc.utils;

import com.huawei.dli.jdbc.model.DliException;
import com.huawei.dli.sdk.common.DLIInfo;

import org.apache.commons.lang3.StringUtils;

import java.net.URI;
import java.nio.charset.Charset;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.SortedMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import lombok.Getter;
import lombok.Setter;

@Getter
@Setter
public class ConnectionResource {

    private static final Pattern KEY_VALUE_REG = Pattern.compile("([^;]*)=([^;]*)[;]?");

    /**
     * keys to retrieve properties from key-value pairs in urls or prop info.
     */
    public static final String ACCESS_KEY_PROP_KEY = "accesskey";

    public static final String ACCESS_KEY_PROP_KEY_1 = "user";

    public static final String SECRET_KEY_PROP_KEY = "secretkey";

    public static final String SECRET_KEY_PROP_KEY_1 = "password";

    public static final String REGION_NAME_PROP_KEY = "regionname";

    public static final String QUEUE_NAME_PROP_KEY = "queuename";

    public static final String ENGINE_TYPE_PROP_KEY = "enginetype";

    public static final String DATABASE_NAME_PROP_KEY = "databasename";

    public static final String SQL_CURRENT_DATABASE_KEY = "dli.sql.current.database";

    public static final String MONITOR_PERIOD_PROP_KEY = "monitorperiod";

    public static final String JOB_TIMEOUT_PROP_KEY = "jobtimeout";

    public static final String PROJECT_ID_PROP_KEY = "projectid";

    public static final String TOKEN_PROP_KEY = "token";

    public static final String LOGCONFFILE_PROP_KEY = "logconffile";

    public static final String CHARSET_PROP_KEY = "charset";

    public static final String DLI_CONF_PREFIX = "dli.";

    public static final String CATALOG = "catalog";

    public static final String SPAKR_CONF_PREFIX = "spark.";

    public static final String CHECK_NO_RESULT_QUERY = "dli.sql.checkNoResultQuery";

    public static final String OBS_ENDPOINT_KEY = "obs.endpoint";

    public static final String USE_CACHE_READER = "read_cache";

    public static final String OBS_ENDPOINT_KEY_1 = "obs_endpoint";

    public static final String OBS_RETRY_INTERVAL = "dli.obs.read.retry.interval.second";

    public static final String LAKEFORMATION_ENDPOINT_KEY = "lakeformation.endpoint";

    public static final String GET_ONE_OBS_RESULT_THREAD_NUM_KEY = "dli.obs.one.result.thread.num";

    public static final String GET_OBS_RESULT_TOTAL_THREAD_NUM_KEY = "dli.obs.result.total.thread.num";

    public static final String SQL_SUBMIT_THREAD_NUM_KEY = "dli.sql.submit.thread.num";

    public static final String RESULT_DATA_LINE_NUM_KEY = "dli.result.line.num";

    public static final String ONE_SQL_RESULT_CACHE_CAPACITY_KEY = "dli.result.queue.cache.capacity";

    // proxy
    public static final String PROXY_HOST = "dli.proxy.host";

    public static final String PROXY_PORT = "dli.proxy.port";

    public static final String PROXY_USER = "dli.proxy.user";

    public static final String PROXY_PASSWORD = "dli.proxy.password";

    // csv read conf
    public static final String CSV_READ_MAX_COLUMN_NUM = "dli.max.column.num";

    // sql conf key
    public static final String DLI_PARTITION_TABLE_FORCE_FILTER_KEY =
        "spark.sql.forcePartitionPredicatesOnPartitionedTable.enabled";

    public static final String DLI_SMALL_MERGE_KEY = "spark.sql.mergeSmallFiles.enabled";

    public static final String DLI_BOOL_TYPE_TO_INT_KEY = "dli.boolean.type.return.int";

    public static final String DLI_DECIMAL_TYPE_TO_STRING_KEY = "dli.decimal.type.return.string";
    
    public static final String DLI_DATE_TYPE_TO_STRING_KEY = "dli.date.type.return.string";

    /**
     * Default value
     */
    private static final String JDBC_LUXOR_URL_PREFIX = "jdbc:luxor://";    // deprecated

    // "jdbc:luxor" will be repreplaced by "jdbc:uquery"
    private static final String JDBC_UQUERY_URL_PREFIX = "jdbc:uquery://";

    private static final String JDBC_DLI_URL_PREFIX = "jdbc:dli://";

    public static final long MONITOR_PERIOD_SECOND = 1L;

    public static final long JOB_TIMEOUT_SECOND = 300L;

    public static final String SERVICE_VERSION = "v1.0";

    public static final String CHARSET_DEFAULT_VALUE = "UTF-8";

    public static final int RESULT_SET_BUFFER_SIZE = 1000;

    // AK/SK required
    private String ak;

    private String sk;

    private String regionName;

    private String host;

    private int port = -1;

    private String queueName;

    private String engineType;

    private String databaseName;

    private String catalog;

    private String projectId;

    private String logConfFile;

    private long monitorPeriod = MONITOR_PERIOD_SECOND;

    private long jobTimeoutSeconds = JOB_TIMEOUT_SECOND;

    private String charSet = CHARSET_DEFAULT_VALUE;

    private boolean checkNoResultQuery = false;

    private String obsEndpoint;

    private boolean useCacheReader = false;

    private int obsRetryInterval;

    private String lakeformationEndpoint;

    private int oneObsResultThreads = 8;

    private int obsResultTotalThreads = 24;

    private int sqlSubmitThreadNum = 8;

    private int resultDataLineNum = -1;

    private int oneSqlQueuesCacheCapacity = 500000;

    private boolean boolTypeToInt = false;

    private boolean decimalTypeToString = false;

    private boolean dateTypeToString = false;

    // proxy
    private String proxyHost;

    private int proxyPort = 80;

    private String userName;

    private String password;

    private String token;

    // csv read conf
    private int maxColumnNum;

    public String getObsEndpoint() {
        if (obsEndpoint == null || obsEndpoint.isEmpty()) {
            return String.format("obs.%s.myhuaweicloud.com", regionName);
        }
        return obsEndpoint;
    }

    public boolean getUserCacheReader() {
        return useCacheReader;
    }

    public String getLakeformationEndpoint() {
        if (lakeformationEndpoint == null || lakeformationEndpoint.isEmpty()) {
            return String.format("lakeformation.%s.myhuaweicloud.com", regionName);
        }
        return lakeformationEndpoint;
    }

    public static boolean acceptURL(String url) {
        return (url != null) && (url.startsWith(JDBC_UQUERY_URL_PREFIX) || url.startsWith(JDBC_LUXOR_URL_PREFIX)
            || url.startsWith(JDBC_DLI_URL_PREFIX));
    }

    public static boolean isCharsetName(String charsetName) {
        SortedMap<String, Charset> charsets = Charset.availableCharsets();
        for (Map.Entry<String, Charset> setEntry : charsets.entrySet()) {
            if (setEntry.getKey().equals(charsetName) || setEntry.getValue().aliases().contains(charsetName)) {
                return true;
            }
        }
        return false;
    }

    public String getEndpoint() {
        return port == -1 ? host : host + ":" + port;
    }

    public static ConnectionResource constructConnectionResource(Properties info, String url)
        throws DliException {
        ConnectionResource connectionResource = extractParamsFromUrl(url);
        // load from property info
        if (info != null) {
            connectionResource.loadPropertiesFromConf(info);
        }
        connectionResource.validateParameters();
        return connectionResource;
    }

    /**
     * Parse JDBC connection URL
     * The format of the URL is:
     * jdbc:dli://&lt;endPoint&gt;/projectId?conf_list
     * jdbc:uquery://&lt;endPoint&gt;/projectId?conf_list is derepcated
     * jdbc:luxor://&lt;endPoint&gt;/projectId?conf_list is derepcated
     * where the conf_list is semicolon separated &lt;key&gt;=&lt;val&gt; pairs.
     * the endPoint can be a domain name or &lt;host&gt;:&lt;port&gt;
     * examples
     * jdbc:dli://localhost:8082/testproject?databasename=db1;queuename=default
     * jdbc:dli://uquery.cn-north-1.myhwclouds.com/testproject?databasename=db1;queuename=default
     *
     * @param url the connection url
     * @return ConnectionResource
     * @throws DliException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public static ConnectionResource extractParamsFromUrl(String url) throws DliException {
        ConnectionResource connectionResource = new ConnectionResource();

        if (!acceptURL(url)) {
            throw new DliException("Bad URL format: Missing prefix " + JDBC_UQUERY_URL_PREFIX,
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }

        // parse the connection uri
        String uriSchema = "schema://";
        int prefixLength;
        if (url.startsWith(JDBC_LUXOR_URL_PREFIX)) {
            prefixLength = JDBC_LUXOR_URL_PREFIX.length();
        } else if (url.startsWith(JDBC_UQUERY_URL_PREFIX)) {
            prefixLength = JDBC_UQUERY_URL_PREFIX.length();
        } else {
            prefixLength = JDBC_DLI_URL_PREFIX.length();
        }
        String uriFromUrl = url.substring(prefixLength);
        URI jdbcURI;
        try {
            jdbcURI = URI.create(uriSchema + uriFromUrl);
        } catch (IllegalArgumentException ex) {
            throw new DliException(String.format("'%s' should follow the URI 'RFC 2396' standard", uriFromUrl),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
        connectionResource.setHost(jdbcURI.getHost());
        connectionResource.setPort(jdbcURI.getPort());

        // get projectid
        String uriPath = jdbcURI.getPath();
        if ((uriPath != null) && !uriPath.isEmpty()) {
            // removing leading '/' returned by getPath()
            uriPath = uriPath.substring(1);
            connectionResource.setProjectId(uriPath);
        }

        // parse key-value pairs
        Map<String, String> parameters = new LinkedHashMap<>();
        String confListStr = jdbcURI.getQuery();
        if (confListStr != null) {
            Matcher confMatcher = KEY_VALUE_REG.matcher(confListStr);
            while (confMatcher.find()) {
                if (parameters.put(confMatcher.group(1), confMatcher.group(2)) != null) {
                    throw new DliException("Bad URL format: Multiple values for property " + confMatcher.group(1),
                        ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            }
        }
        connectionResource.loadPropertiesFromConf(parameters);
        return connectionResource;
    }

    public void loadPropertiesFromConf(Map map) throws DliException {
        if (this.projectId == null && map.get(PROJECT_ID_PROP_KEY) != null) {
            this.setProjectId((String) map.get(PROJECT_ID_PROP_KEY));
        }
        String accessKey = (String) map.get(ACCESS_KEY_PROP_KEY);
        String userNameAccessKey = (String) map.get(ACCESS_KEY_PROP_KEY_1);
        String secretKey = (String) map.get(SECRET_KEY_PROP_KEY);
        String passwordSecretKey = (String) map.get(SECRET_KEY_PROP_KEY_1);

        if (StringUtils.isNoneBlank(accessKey, secretKey)) {
            ak = accessKey;
            sk = secretKey;
        } else if (StringUtils.isNoneBlank(userNameAccessKey, passwordSecretKey)) {
            ak = userNameAccessKey;
            sk = passwordSecretKey;
        }

        if (map.get(REGION_NAME_PROP_KEY) != null) {
            this.setRegionName((String) map.get(REGION_NAME_PROP_KEY));
        }
        if (map.get(QUEUE_NAME_PROP_KEY) != null) {
            this.setQueueName((String) map.get(QUEUE_NAME_PROP_KEY));
        }
        if (map.get(ENGINE_TYPE_PROP_KEY) != null) {
            this.setEngineType((String) map.get(ENGINE_TYPE_PROP_KEY));
        }
        if (map.get(DATABASE_NAME_PROP_KEY) != null) {
            this.setDatabaseName((String) map.get(DATABASE_NAME_PROP_KEY));
        }
        if (map.get(SQL_CURRENT_DATABASE_KEY) != null) {
            this.setDatabaseName((String) map.get(SQL_CURRENT_DATABASE_KEY));
        }
        if (map.get(MONITOR_PERIOD_PROP_KEY) != null) {
            try {
                long period = Long.parseLong((String) map.get(MONITOR_PERIOD_PROP_KEY));
                if (period > 0) {
                    this.setMonitorPeriod(period);
                } else {
                    throw new DliException(
                        String.format("%s can only be an integer greater than 0", MONITOR_PERIOD_PROP_KEY),
                        ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            } catch (NumberFormatException ex) {
                throw new DliException(
                    String.format("%s can only be an integer greater than 0", MONITOR_PERIOD_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(JOB_TIMEOUT_PROP_KEY) != null) {
            try {
                long jobTimeOut = Long.parseLong((String) map.get(JOB_TIMEOUT_PROP_KEY));
                if (jobTimeOut > 0) {
                    this.setJobTimeoutSeconds(jobTimeOut);
                } else {
                    throw new DliException(String.format("%s can only be an integer greater than 0",
                        JOB_TIMEOUT_PROP_KEY), ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            } catch (NumberFormatException ex) {
                throw new DliException(String.format("%s can only be an integer greater than 0", JOB_TIMEOUT_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(LOGCONFFILE_PROP_KEY) != null) {
            this.setLogConfFile((String) map.get(LOGCONFFILE_PROP_KEY));
        }
        if (map.get(CHARSET_PROP_KEY) != null) {
            String charset = (String) map.get(CHARSET_PROP_KEY);
            if (isCharsetName(charset)) {
                this.setCharSet(charset);
            } else {
                throw new DliException(String.format("The charset '%s' is not supported!", charset),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(TOKEN_PROP_KEY) != null) {
            this.setToken((String) map.get(TOKEN_PROP_KEY));
        }
        if (map.get(CHECK_NO_RESULT_QUERY) != null) {
            this.setCheckNoResultQuery(Boolean.parseBoolean((String) map.get(CHECK_NO_RESULT_QUERY)));
        }
        if (map.get(OBS_ENDPOINT_KEY) != null) {
            this.obsEndpoint = (String) map.get(OBS_ENDPOINT_KEY);
        }
        if (map.get(USE_CACHE_READER) != null) {
            this.useCacheReader = (Boolean.parseBoolean((String) map.get(USE_CACHE_READER)));
        }
        if (map.get(OBS_ENDPOINT_KEY_1) != null) {
            this.obsEndpoint = (String) map.get(OBS_ENDPOINT_KEY_1);
        }
        if (map.get(OBS_RETRY_INTERVAL) != null) {
            this.obsRetryInterval = Integer.parseInt((String) map.get(OBS_RETRY_INTERVAL));
        }
        if (map.get(LAKEFORMATION_ENDPOINT_KEY) != null) {
            this.lakeformationEndpoint = (String) map.get(LAKEFORMATION_ENDPOINT_KEY);
        }
        if (map.get(GET_ONE_OBS_RESULT_THREAD_NUM_KEY) != null) {
            this.oneObsResultThreads = Integer.parseInt((String) map.get(GET_ONE_OBS_RESULT_THREAD_NUM_KEY));
        }
        if (map.get(GET_OBS_RESULT_TOTAL_THREAD_NUM_KEY) != null) {
            this.obsResultTotalThreads = Integer.parseInt((String) map.get(GET_OBS_RESULT_TOTAL_THREAD_NUM_KEY));
        }
        if (map.get(SQL_SUBMIT_THREAD_NUM_KEY) != null) {
            this.sqlSubmitThreadNum = Integer.parseInt((String) map.get(SQL_SUBMIT_THREAD_NUM_KEY));
        }
        if (map.get(ONE_SQL_RESULT_CACHE_CAPACITY_KEY) != null) {
            this.oneSqlQueuesCacheCapacity = Integer.parseInt((String) map.get(ONE_SQL_RESULT_CACHE_CAPACITY_KEY));
            if (this.oneSqlQueuesCacheCapacity <= 0) {
                this.oneSqlQueuesCacheCapacity = Integer.MAX_VALUE;
            }
        }
        if (map.get(RESULT_DATA_LINE_NUM_KEY) != null) {
            this.resultDataLineNum = Integer.parseInt((String) map.get(RESULT_DATA_LINE_NUM_KEY));
        }
        if (map.get(DLI_BOOL_TYPE_TO_INT_KEY) != null) {
            this.boolTypeToInt = Boolean.parseBoolean((String) map.get(DLI_BOOL_TYPE_TO_INT_KEY));
        }
        if (map.get(DLI_DECIMAL_TYPE_TO_STRING_KEY) != null) {
            this.decimalTypeToString = Boolean.parseBoolean((String) map.get(DLI_DECIMAL_TYPE_TO_STRING_KEY));
        }
        if (map.get(DLI_DATE_TYPE_TO_STRING_KEY) != null) {
            this.dateTypeToString = Boolean.parseBoolean((String) map.get(DLI_DATE_TYPE_TO_STRING_KEY));
        }
        if (map.get(CATALOG) != null) {
            this.catalog = (String) map.get(CATALOG);
        }
        if (map.get(PROXY_HOST) != null) {
            this.proxyHost = (String) map.get(PROXY_HOST);
        }
        if (map.get(PROXY_PORT) != null) {
            this.proxyPort = Integer.valueOf((String) map.get(PROXY_PORT));
        }
        if (map.get(PROXY_USER) != null) {
            this.userName = (String) map.get(PROXY_USER);
        }
        if (map.get(PROXY_PASSWORD) != null) {
            this.password = (String) map.get(PROXY_PASSWORD);
        }
        if (map.get(CSV_READ_MAX_COLUMN_NUM) != null) {
            this.maxColumnNum = Integer.parseInt((String) map.get(CSV_READ_MAX_COLUMN_NUM));
        }
    }

    private void validateParameters() throws DliException {
        String errorMsg = "The configuration item '%s' must be set in AK/SK authentication mode";
            if (StringUtils.isBlank(ak) || StringUtils.isBlank(sk)) {
                throw new DliException(String.format(errorMsg, ACCESS_KEY_PROP_KEY + "/" + SECRET_KEY_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
            if (regionName == null || regionName.isEmpty()) {
                throw new DliException(String.format(errorMsg, REGION_NAME_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }

        if (projectId == null || projectId.isEmpty()) {
            throw new DliException(String.format("%s must be set", PROJECT_ID_PROP_KEY),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
        if (queueName == null || queueName.isEmpty()) {
            throw new DliException(String.format("%s name must be set", QUEUE_NAME_PROP_KEY),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
    }

    public DLIInfo toDliInfo() {
        DLIInfo dliInfo = new DLIInfo(getRegionName(), getAk(), getSk(), getProjectId());
        dliInfo.setDliEndpoint(getEndpoint());
        dliInfo.setLfEndpoint(getLakeformationEndpoint());
        dliInfo.setObsEndpoint(getObsEndpoint());
        dliInfo.setQueueName(getQueueName());
        dliInfo.setProxyHost(getProxyHost());
        dliInfo.setProxyPort(getProxyPort());
        dliInfo.setProxyUser(getUserName());
        dliInfo.setProxyPwd(getPassword());
        dliInfo.setUseRetry(true);
        dliInfo.setUseCacheReader(useCacheReader);
        return dliInfo;
    }
}
